// ClkISR.c -- Dutchtronix ISR Support
//
//  Copyright  2010 Johannes P.M. de Rie
//
//  All Rights Reserved
//
//  This file is part of the Dutchtronix Oscilloscope Clock Distribution.
//  Use, modification, and re-distribution is permitted subject to the
//  terms in the file named "LICENSE.TXT", which contains the full text
//  of the legal notices and should always accompany this Distribution.
//
//  This software is provided "AS IS" with NO WARRANTY OF ANY KIND.
//
//  This notice (including the copyright and warranty disclaimer)
//  must be included in all copies or derivations of this software.
//
#include <avr/io.h>
#include <avr/portpins.h>
#include <stdlib.h>
#include <string.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>
#include <avr/wdt.h>

#include "./config.h"
#include "./ClkISR.h"
#include "./sound.h"

extern void InitRTCInt(void);
extern void ResetPPSCntr(void);		//TODO should be macro or inline function
extern void FuncGenProcessInput(void);

PushButtonState_t S1;
PushButtonState_t S2;
PushButtonState_t S3;
PushButtonState_t S4;


#if DEBUG
byte	TC0DebugCnt;
#endif

//
// TC0 is used to program the push buttons
//
// It tracks the duration of a down push and it does key debouncing
//
// Initialize TC0
//
// Set prescaler to clkio/1024;
//	At 20 Mhz, this is 51.2 uSec. Timer0CountUp value 127 -> 6.502 mSec
//	At 16 Mhz, this is 64 uSec.	Timer0CountUp value 101 -> 6.464 mSec
//
void InitTC0(void)
{
	TCCR0A = 0;
	TCCR0B = (1<<CS02)|(0<<CS01)|(1<<CS00);
	OCR0A = Timer0CountUp;
}
//
// Enable TC0 timer for key debouncing
//
void StartKeyDebouncing(void)		//EnableTC0
{
	GPIOR0 |= _BV(fDebouncing);
	S1.TC0Cntr = 0;
	S1.KeyState = 0xff;
	S1.Position = FALSE;
	S1.fAutorepeat = FALSE;
	S2 = S1;
	S3 = S1;
	S4 = S1;
#if DEBUG
	TC0DebugCnt = 150;
#endif
	TIFR0 = _BV(OCF0A);				//clear pending match
	TIMSK0 = _BV(OCIE0A);			//Enable TC0.matchA
}
//
// Disable TC0 timer for key debouncing
//
void DisableTC0(void)
{
	TIMSK0 = (0<<OCIE0A);			//Disable TC0.matchA
}
//
void ProcessButton(PushButtonState_t *pSW, byte ButtonMsk, byte fUPMsk, byte fTimeoutMsk, BOOL fDemoEnabled)
{
//
// Decrement TC0Cntr iff != 0
//
	if (pSW->TC0Cntr && (--pSW->TC0Cntr == 0)) {
		//
		// the 8-bit Timer0S1Counter field is now 0
		//
		if (!fDemoEnabled) {							//ignore if in demo mode
			GPIOR1 |= fTimeoutMsk;				//mark TC0 timeout in flags
			pSW->Position = FALSE;						//Pretend Button is up (ignore next fUPS1)
			pSW->fAutorepeat = FALSE;
		}
	}
//
// bool_t DebounceSwitch2() Jack Ganssle embedded.com
//    static uint16_t State = 0// // Current debounce status
//    State=(State<<1) | !RawKeyPressed() | 0xe000//
//    if(State==0xf000)return TRUE//
//    return FALSE// 
//
	byte v = SWPin & ButtonMsk;
	if (v) v = 0b0000001;
	pSW->KeyState = (uint8_t)(pSW->KeyState << 1) | v;		//0 bit value means down
	if (pSW->KeyState == 0b10000000) {						//7 downs in a row
		if (pSW->Position == FALSE) {
			pSW->Position = TRUE;
			pSW->TC0Cntr = (TIMER0TIMEOUT & 0xff);
		}
	} else if (pSW->KeyState == 0b01111111) {				//7 ups in a row. Could be Button Up or failed push (<7 downs in a row)
		if (pSW->fAutorepeat) {
			pSW->fAutorepeat  = FALSE;
			GPIOR1 &= ~fTimeoutMsk;
		} else if (pSW->Position != FALSE) {				//If Button already UP (FALSE), ignore this event
			GPIOR0 |= fUPMsk;
		}
		pSW->Position = FALSE;
		pSW->TC0Cntr = 0;
	} else if (pSW->KeyState == 0) {
		//
		// pSW->KeyState==0 indicates key has been down for at least 8 checks.
		// Start autorepeating (if allowed) after S1 has had a timeout (fTimeoutS1)
		//
		if (pSW->okToRepeat && (GPIOR1 & fTimeoutMsk)) {
			if (pSW->fAutorepeat == FALSE) {
				pSW->Position = TRUE;		//Undo "Pretend Button is up" (we want the next fUPS1)
				pSW->autorepeatcnt = AUTOREPEATVAL;
				pSW->fAutorepeat = TRUE;
			} else {
				if (--pSW->autorepeatcnt == 0) {
					//
					// reached 0. Insert an fUPS1 event.
					//
					GPIOR0 |= fUPMsk;	//set UP in flags
					pSW->autorepeatcnt = AUTOREPEATVAL;
				}
			}
		}
	}
}
//
// Timer0 Match Handler. Used to debounce/interpret S1 and S2
//
ISR(TIMER0_COMPA_vect)
{
	TCNT0 = 0;
	sei();											//stay as accurate as possible

	ProcessButton(&S1, _BV(SW1Bit), _BV(fUPS1), _BV(fTimeoutS1), FALSE);
	ProcessButton(&S2, _BV(SW2Bit), _BV(fUPS2), _BV(fTimeoutS2), FALSE);
	ProcessButton(&S3, _BV(SW3Bit), _BV(fUPS3), _BV(fTimeoutS3), FALSE);
	ProcessButton(&S4, _BV(SW4Bit), _BV(fUPS4), _BV(fTimeoutS4), FALSE);

	if ((S1.KeyState & S2.KeyState & S3.KeyState & S4.KeyState) == 0xff) {		//all ones means UP. done if all switches are all ones
		GPIOR0 &= ~_BV(fDebouncing);
		DisableTC0();								//stop key debouncing timer
	}
}
//
//
void EnableTC3(void)
{
	TIMSK3 |= _BV(OCIE3A);				//Enable TC1.matchA
	Timer3Counter = 0;
}
//
//
void DisableTC1(void)
{
	TIMSK3 &= ~_BV(OCIE3A);				//Disable TC1.matchA
}
//
// TC3 is the general purpose timer
//
// We use the 16-bit timer for this purpose since it will
// get us closest to precisely 1 second.
//
// Set prescaler to clkio/256.
//	 At 20 Mhz, this is 12.8 uSec 
//	 At 16 Mhz, this is 16 uSec
//
// set TC3 to interrupt at 60 Hz.
//
void InitTC3(void)
{
	SecondsCnt = ReloadSecondsCnt;		//initialize tick counter for timeout
	TickCnt = 0;
	TCCR3B = 0b00000100;				//TC1.ck = 20Mhz/256 = 12.8 uSec
	OCR3A = Timer3CountUp;				// or 16 Mhz/256 = 16 uSec
	EnableTC3();
}
//
// SetTC1Countdown(uint v)
//
void SetTC3Countdown(uint v)
{
	byte flags = SREG;
	cli();
	Timer3Counter = v;
	SREG = flags;
}
//
// TC1 countdown used to timeout after 30 seconds.
// an interrupt could occur while doing the assignment (2 instructions)
//
void StartTC3Countdown(void)
{
	SetTC3Countdown(TIMER3TIMEOUT);
}
//
// return 0 if counter is NULL, else return 1
//
byte TestTC3Countdown(void)
{
	return (Timer3Counter != 0);
}
//
// Timer3 Compare Match Handler.
//
// Operates at 60 Hz (16.666 mSecs)
//
// Use the 16 bit counter since it approaches 1 second closest.
//
// Other timers to be added (1 byte counter yields max 4.2 seconds delay)
//
ISR(TIMER3_COMPA_vect)
{
	TCNT3 = 0;
	++TickCnt;
//
// Check if the push button switch S1 is closed.
// Skip if we are in debouncing mode
// if HIGH (set), button is UP. If LOW (clear), button is DOWN
//
	if (!(GPIOR0 & _BV(fDebouncing))) {		//skip test if already debouncing
		//check if any push button switch pin is low (meaning down)
		if ((SWPin	& (_BV(SW1Bit)|_BV(SW2Bit)|_BV(SW3Bit)|_BV(SW4Bit))) != (_BV(SW1Bit)|_BV(SW2Bit)|_BV(SW3Bit)|_BV(SW4Bit))) {
			StartKeyDebouncing();
		}
	}
#if 0
//
// Check if the push button switch S2 is closed.
// Reboot if it is.
//
	if (GPIOR0 & _BV(fTimeoutS2)) {					//did SW2 timeout
		GPIOR0 &= ~_BV(fTimeoutS2);
		InitiateSysReset();
	}
#endif
//
// Decrement Timer3Counter iff != 0
//
	if (Timer3Counter && (--Timer3Counter == 0)) {
		//
		// the 16-bit Timer3Counter field is now 0
		//
	}
//
//  Seconds counter
//
	if (SecondsCnt && (--SecondsCnt == 0)) {
		GPIOR0 |= _BV(fSecond);				//set TimeOut Flag
		SecondsCnt = ReloadSecondsCnt;		//reinitialize tick counter for timeout
	}
}

void InitLed(void)
{
	LEDDDR |= (_BV(LEDBit0) | _BV(LEDBit1));			//set as output
	LEDPort &= ~(_BV(LEDBit0) | _BV(LEDBit1));			//turn off
}
//
void BlinkLed0(void)
{
	LEDPin |= _BV(LEDBit0);				//toggle the LED
}
//
void BlinkLed1(void)
{
	LEDPin |= _BV(LEDBit1);				//toggle the LED
}
//
void TurnOffLed(void)
{
	LEDPort &= ~_BV(LEDBit0);
}

#if DEBUG
void DBGTurnLedOn(void)
{
	LEDPort |= _BV(LEDBit0);
}
//
void DBGTurnLedOff(void)
{
	LEDPort &= ~_BV(LEDBit0);
}
#endif

//
void InitSwitch(void)
{
	SWDDR &= ~(_BV(SW1Bit)|_BV(SW2Bit)|_BV(SW3Bit)|_BV(SW4Bit));	//set as input
	SWPort |= (_BV(SW1Bit)|_BV(SW2Bit)|_BV(SW3Bit)|_BV(SW4Bit));	//enable pull-up resistor
}
//
//	Execute a read on the EEPROM
//	arg: eeprom address to read
//	out: data byte
//	no address increment
//
byte EepromRead(uint adr)
{
	do {} while (EECR & _BV(EEPE));
	EEAR = adr;
	EECR |= _BV(EERE);							//read command
	return EEDR;
}
//
// EepromWriteNew(uint EEpromadr, byte val)
//
void EepromWrite(uint adr, byte val)
{
	do {} while (EECR & _BV(EEPE));
	EEAR = adr;
	EEDR = val;
	cli();
	EECR |= _BV(EEMPE);							//write command
	EECR |= _BV(EEPE);
	sei();
}
//
// Disable Watchdog Timer Completely
//
void WDTimerOff(void)
{
	wdt_reset();
	MCUSR &= ~_BV(WDRF);
	WDTCSR |= (_BV(WDCE) | _BV(WDE));
	WDTCSR = 0;
}
//
// Initiate a System Reset
//
void InitiateSysReset(void)
{
	cli();
	wdt_reset();
	WDTCSR |= (_BV(WDCE) | _BV(WDE));
	WDTCSR = (_BV(WDE)|_BV(WDP1));				//Watchdog System Reset Enable, 64 mSec timeout
	sei();
	do {} while (TRUE);
}

#define LF 10

#define MAXTXBUFFERLEN	256
#define MAXRXBUFFERLEN	128
#define TXBUFFERMASK	0b11111111	//2^8-1
#define RXBUFFERMASK	0b01111111	//2^7-1

#if 1
char	TXBuffer[MAXTXBUFFERLEN];
char	RXBuffer[MAXRXBUFFERLEN];
volatile BOOL UARTOutReady;
volatile byte TXBufferHead;
volatile byte TXBufferTail;
volatile byte RXBufferHead;
volatile byte RXBufferTail;
#endif

extern void BlinkLed1(void);

void UARTInit(uint newBaudVal, BOOL fUSE_2X)
{
	UCSR0A = 0;
	if (fUSE_2X) {
		UCSR0A |= _BV(U2X0);
	} else {
		UCSR0A &= ~_BV(U2X0);
	}
// set rx interrupt, tx interrupt, clear data empty interrupt
	UCSR0B = _BV(TXCIE0)|_BV(TXEN0)|_BV(RXCIE0)|_BV(RXEN0);
// set default 8 bits, no parity, 1 stop
	UCSR0C = (_BV(UCSZ01)|_BV(UCSZ00));
	UARTOutReady = TRUE;
	UBRR0 = newBaudVal;

#if DEBUG
//
// all memory cleared in RELEASE mode
//
	RXBufferHead = 0;
	RXBufferTail = 0;
#if SCOPECLOCK
	RXLFDetected = 0;
#endif
	TXBufferHead = 0;
	TXBufferTail = 0;
#endif
}
//
//
#if SCOPECLOCK
uint MapBaudVal2BaudRate(byte bVal)
{
	return pgm_read_word(&MapBaudVal[bVal]);
}
#endif
//
//
void UARTSetBaudVal(uint newBaudVal)
{
	byte sreg = SREG;
	cli();
	UBRR0 = newBaudVal;
	SREG = sreg;
}

#if 0
void UART_Set4800(void)
{
	UBRR0 = 259;					//baud 4800 at 20MHz Xtal 
}

void UART_Set19200(void)
{
	UBRR0 = 64;						//baud 19200 at 20MHz Xtal 
}
#endif
//
// Change Baud Rate to 115,200 bps
//
void UARTSet115200(void)
{
	UBRR0 = 10;						//baud 115200 at 20MHz Xtal
}

void UARTStop(void)
{
	UCSR0A = 0;
	UCSR0B = 0;
	UCSR0C = 0;
	UARTOutReady = FALSE;
}
//
// UART Transmit Complete Interrupt
//
ISR(USART0_TX_vect)
{
	byte tail = TXBufferTail;				//remove volatile attribute since ints are off
	if (TXBufferHead == tail) {
		UARTOutReady = TRUE;				// no more buffered data
		return;
	}
	UDR0 = TXBuffer[tail];
	TXBufferTail = ((tail + 1) & TXBUFFERMASK);
}
//
// UART Receive Complete Interrupt
//
ISR(USART0_RX_vect)
{
	byte newHead = (RXBufferHead + 1) & RXBUFFERMASK;
	byte b = UDR0;
	if (RXBufferTail == newHead) {
		// no room for incoming character in RingBuffer
#if SCOPECLOCK
		// TODO bug if non-clock mode
		if (EEConfigData.GPSInEnabled) {
			RXLFDetected = TRUE;	// make sure we empty the buffer
		}
#endif
	} else {
		RXBuffer[RXBufferHead] = b;
		GPIOR0 |= _BV(fRcvdChar);
#if SCOPECLOCK
		if (RXBuffer[RXBufferHead] == LF) {
			if (EEConfigData.GPSInEnabled) {
				RXLFDetected = TRUE;	// make sure we empty the buffer
			}
		}
#endif
		RXBufferHead = newHead;
	}
#if SCOPECLOCK
	// use TC1 to force FuncGen to process the input
	if ((G3_FLAGS & _BV(fMiniDDS)) && !(G3_FLAGS & _BV(fInputActive))) {
		TIMSK1 = _BV(OCIE1A);		//Enable TC1.Match A
	}
#endif
}
//
// void UARTSendByte(char ch, BOOL fForce)
//
void UARTSendByte(byte b, BOOL fForce)
{
#ifndef TERMOUTONLY
	if (UARTOutReady) {					//if ready, send out
		UARTOutReady = FALSE;
		UDR0 = b;
	} else {
		byte newHead = (TXBufferHead + 1) & TXBUFFERMASK;
		if (TXBufferTail == newHead) {		//if equal, there is no room
			if (fForce) {
				do {
#if SCOPECLOCK
					fpCurrentRefresh();											//draw the display.
#endif
				} while (UARTOutReady == FALSE);
				BlinkLed1();
				UARTOutReady = FALSE;
				UDR0 = b;
			}
		} else {
			TXBuffer[TXBufferHead] = b;
			TXBufferHead = newHead;
		}
	}
#endif
}
//
// Print String
//
void UARTPrintfStr(char *pBuf)
{
	char ch;
	while ((ch = *pBuf++) != 0) {
		UARTSendByte(ch, TRUE);
	}
}
//
// Print flash based text
//
void UARTPrintfProgStr(PGM_P pBuf)
{
	char ch;
	while ((ch = pgm_read_byte(pBuf)) != 0) {
		UARTSendByte(ch, TRUE);
		++pBuf;
	}
}
//
// RXBufferHead may change while executing this function
//
byte UARTReceiveByte(void)
{
	byte b;
	while (!(GPIOR0 & _BV(fRcvdChar))) ;
#if SCOPECLOCK
	ASSERT(RXBufferHead != RXBufferTail);
#endif
	b = RXBuffer[RXBufferTail];
	RXBufferTail = (RXBufferTail + 1) & RXBUFFERMASK;
	if (RXBufferHead != RXBufferTail) {
		return b;						// more chars available
	}
	GPIOR0 &= ~_BV(fRcvdChar);
	return b;
}
//
#if SCOPECLOCK
//
BOOL UARTReceiveLine(char *pDst, byte maxLength)
{
	char ch;
	byte head, tail, l = 0;
	byte sreg = SREG;
	cli();
	ASSERT(GPIOR0 & _BV(fRcvdChar));
	head = RXBufferHead;
	tail = RXBufferTail;
	do {
		ch = RXBuffer[tail];
		tail = (tail + 1) & RXBUFFERMASK;
		if (++l < maxLength) {
			*pDst++ = ch;
		}
	} while ((ch != LF) && (head != tail));
	if (head == tail) {					//is buffer now empty?
		GPIOR0 &= ~_BV(fRcvdChar);		//buffer now empty
		*--pDst = LF;					//make sure there is a LF
	}
	RXBufferTail = tail;
	SREG = sreg;
	return l;
}
//
// Flush remainder of current record (until LF or end)
//
void UARTFlushRecord(void)
{
	char ch;
	byte head, tail;
	if (GPIOR0 & _BV(fRcvdChar)) {
		byte sreg = SREG;
		cli();
		head = RXBufferHead;
		tail = RXBufferTail;
		do {
			ch = RXBuffer[tail];
			tail = (tail + 1) & RXBUFFERMASK;
		} while ((ch != LF) && (head != tail));
		if (head == tail) {						//is buffer now empty?
			GPIOR0 &= ~_BV(fRcvdChar);			//yes
		}
		RXBufferTail = tail;
		SREG = sreg;
	}
}
#endif
//
//
void UARTFlushPendingOutput(void)
{
	TXBufferHead = 0;
	TXBufferTail = 0;
}

typedef struct
{
    // Maximum an Ticks fr Timer ISR bis zum Fertigstellen von 
    // pixel_on_the_fly()
    uint16_t ticks_max;
    uint8_t loop;
} statistic_t;

typedef struct
{
    volatile uint8_t debug __attribute__((deprecated));
    uint8_t bootlevel;
    uint8_t mcusr;
    statistic_t stat;
    
    // Timer-ISR
    struct
    {
        uint16_t num_10ms;
        uint8_t  num_1ms;
        uint8_t  ocr_err;
    } irq;
    
} system_t;

system_t sys;

#define PRESCALE 8
#define IRQS 48000

static const uint8_t fcpu_err[] PROGMEM = 
{
#if (F_CPU % (PRESCALE * IRQS)) * 2 == PRESCALE * IRQS
    #define ERR_NUM    1
    #define ERR_CYCLES 2
#elif (F_CPU % (PRESCALE * IRQS)) * 3 == PRESCALE * IRQS
    #define ERR_NUM    1
    #define ERR_CYCLES 3
#elif (F_CPU % (PRESCALE * IRQS)) * 3 == 2 * PRESCALE * IRQS
    #define ERR_NUM    2
    #define ERR_CYCLES 3
#else
#if (F_CPU % (PRESCALE * IRQS)) != 0 
#   warning Timer2 arbeitet ungenau
#endif
    #define ERR_NUM    0
    #define ERR_CYCLES 0
#endif
};

//#undef ERR_CYCLES
//#define ERR_CYCLES 0

ISR (TIMER2_COMPA_vect)
{
    while (1)
    {
        wdt_reset();

        // Fehler aufgrund nicht-passener IRQ-Rate ausbgeln, d.h.
        // wenn F_CPU % (PRESCALE * IRQs) != 0 ist. Der verbleibenede
        // Rest wird in PWM-hnlicher Manier auf die IRQ-Rate aufgeschlagen.
        if (ERR_CYCLES == 2)
        {
            uint8_t ocr = (uint32_t) F_CPU / PRESCALE / IRQS-1;
            int8_t tcnt2 = TCNT2-ocr;
            if (tcnt2 <= -3 || tcnt2 >= 2)
            {
                ocr = 2 * ocr+1;
                OCR2A = ocr - OCR2A;
            }
        }
/*        
        if (ERR_CYCLES)
        {
            static uint8_t isr_error;
            uint8_t err = 1 + isr_error;

            if (ERR_CYCLES == 2)
            {
                err %= 2;
                ocr += err;
            }
            else 
            {    
                if (err <= ERR_NUM)           ocr++;
                if (err >= ERR_CYCLES)        err = 0;
            }
            
            isr_error = err;

            OCR2A = ocr;
        } // ERR_CYCLES
*/            
#if 0
        pixel_on_the_fly ();
        if (sys.irq.num_10ms % 2 == 1)
            rc5_poll();
        
#if FB_REG && __GNUC__ < 4 && defined __AVR__
        asm volatile ("; clobber Y":::"28","29");
#endif
#endif
        ////////////////////////////////////////////////////////////
        // 1 ms-Takt fr onRedraw
        uint8_t irqs_1ms = 1 + sys.irq.num_1ms;
        if (irqs_1ms >= IRQS / 1000)
        {
            // 1 ms ist voll
//            menu.tick_1ms = 1;
            irqs_1ms = 0;
        }
        sys.irq.num_1ms = irqs_1ms;	
        
        ////////////////////////////////////////////////////////////
        // 10 ms-Takt fr onRedraw
        // 10 ms-Takt fr Jobs: countdown-Zhler, Tasten-Entprellung, DCF, ...
        uint16_t irqs_10ms = 1 + sys.irq.num_10ms;
        if (irqs_10ms >= IRQS / 100)
        {
            // 10 ms sind voll
//            menu.tick_10ms = 1;
            irqs_10ms = 0;
        }
            
        sys.irq.num_10ms = irqs_10ms;	
        
        if (irqs_10ms <= 7)
        {
            uint8_t i = irqs_10ms;

            // Immer Lcken zum Erholen lassen
//			if (i == 1)   job_countdown();
//            if (i == 3)   job_morse (&morse_data);
            if (i == 5)   job_sound (&sdata);
//            if (i == 7)   job_dcf ();
        }
        
        // Direkt zurckspringen spart einen ISR-Epilog und -Prolog,
        // die zusammen rund 90 Ticks kosten wrden!
        if (TIFR2 & (1 << OCF2A))
        {
            TIFR2 = (1 << OCF2A);
            continue;
        }
        
        if (TCNT2 < 56)
            return;
            
        while (! (TIFR2 & (1 << OCF2A)));
        TIFR2 = (1 << OCF2A);
        
        //while (TCNT2 < 5);
    }
}

/*****************************************************************************************/
/*****************************************************************************************/
/*****************************************************************************************/

void timer2_init (void)
{
    // Mode #2 (CTC) und PRESCALE=8
    TCCR2A = (1 << WGM21);
    TCCR2B = (1 << CS21);

    // OutputCompare fr gewnschte Timer2 Frequenz
    OCR2A = (unsigned short) ((unsigned long) F_CPU / PRESCALE / IRQS-1);

    // OutputCompare-Interrupt A fr Timer 2
    TIMSK2 = (1 << OCIE2A);		//xxxx
}

