; DSTime.asm - included source for Dutchtronix Oscilloscope Clock
;
;  Copyright  2008 Johannes P.M. de Rie
;
;  All Rights Reserved
;
;  This file is part of the Dutchtronix Oscilloscope Clock Distribution.
;  Use, modification, and re-distribution is permitted subject to the
;  terms in the file named "LICENSE.TXT", which contains the full text
;  of the legal notices and should always accompany this Distribution.
;
;  This software is provided "AS IS" with NO WARRANTY OF ANY KIND.
;
;  This notice (including the copyright and warranty disclaimer)
;  must be included in all copies or derivations of this software.
;
; Firmware for Dutchtronix AVR Oscilloscope Clock
;
; after reading time and date from RTC, get DSTForward and DSTBack for current year
;
;
; DST jump will be missed if clock is not running at transition.
; Flags to check at startup time:
; Bitvector for 7 years: was forward DST applied, was backward DST applied.
; If DSTYear and date/time window and Forward flag not set  >= Forward date/time, inc hr, may overflow date!
; If DSTYear and after date/time window and Forward flag set and Backward flag NOT set ->
;	back one hour. May set date back!
; flags cleared if date before Forward date
; Year change-> redo DST data
;
; Source: http://webexhibits.org/daylightsaving/b.html
; Consistency check: if date/time is prewindow and DST Forw was applied?
; If time is set and we are INSIDE a DST window, mark DST Forward as set
; It time is set and we are AFTER a DST window, mark DST Backward as set
;
; Regular DST correction is kind of useless when using NMEA data to set the clock
;
; Time was changed by the user (Menu or Serial Interface).
; Do NOT apply a DST correction (if applicable) but mark the bitvectors
; that a correction was applied (if applicable).
; The theory here is that the user knows what he or she is doing.
;
DSTRetroNotApply:
	sts		AllowDSTFix,_0		;mode flag
	rjmp	DSTRetro
;
; DSTRetroApply--
;	After the clock starts up, or after the user changed the date, check
;	if the new date requires a DST fixup.
;
DSTRetroApply:
	sts		AllowDSTFix,_1		;mode flag
DSTRetro:
;
;initialize Daylight Saving Time variables
;
	ldiw	Z,AVRTimeDateBlk
	rcall	DSTInitfromZ		;preserves Z
	rcall	DSTInWindowfromZ	;preserves Z
	brcs	DSTRA100			;not in window
;
; Time should have been advanced. Was it done?
;	
	rcall	DSTFWasApplied		;applied to current year? Preserves Z
	tst		r16
	brne	DSTRA800			;Yes, done
;
; Advance time one hour. May change date. Cannot change year.
; Same issue with GPS input if time is in the last hr of the window.
; time will be advanced one hour (but only once), then set back at the next minute.
;
	lds		r16,AllowDSTFix		;Fix allowed?
	sbrc	r16,0
	CALL	AdvanceAVRTimeOneHr	;expects Z to be loaded
	rcall	DSTFMarkApplied
;
; Time may have advanced beyond the DST window (if we were in the last hour of the window)
; and we were allowed to apply a fix. Check again to see if we have to apply a backward
; fix. Recursion is ok here since the Forward bitvector was set, implying we won't do
; another forward fix (and check again etc)
;
	rcall	DSTRetroApply		;recurse to see if we advanced beyond the DST window.
	rjmp	DSTRA800			;done
;
; check if the clock is after the DST window and DST Forward was applied
;
DSTRA100:
;	ldiw	Z,AVRTimeDateBlk	;not needed
	rcall	DSTPostWindowfromZ
	brcs	DSTRA800			;not in window
	rcall	DSTFWasApplied		;applied to current year? Preserves Z
	tst		r16
	breq	DSTRA800			;No, done
;
; Time should have been put back. Was it done?
;	
	rcall	DSTBWasApplied		;applied to current year? Preserves Z
	tst		r16
	brne	DSTRA800			;Yes, done
;
;	put back time one hour. May change date. Cannot change year
;
	lds		r16,AllowDSTFix		;Fix allowed?
	sbrc	r16,0
	CALL	RetreatAVRTimeOneHr	;expects Z to be loaded
	rcall	DSTBMarkApplied
DSTRA800:
	ret							;done
;
; Check is DST is enabled and if the year in the time block
;	pointed to by Z is a supported year
;
; Supported year: 2008-2014 (7 years. Can't do 8 years with bitvectors in EEProm)
;
; OUT:	CC yes, CS no
;		r16 - normalized year (base 2008)
;
; Preserves Z
;
DSTPossiblefromZ:
	lds		r16,DSTMode		;0..2 are valid values
	cpi		r16,DSTNONE		;OFF Value
	breq	DSTnodata
	ldd		r16,pYear		;supported year?
	subi	r16,8			;normalize
	brcs	DSTnodata
	cpi		r16,8			;max year + 1
	brsh	DSTnodata
	clc
	ret
DSTnodata:
	sec
	ret
;
; Check is DST is enabled and if the current AvrYear is a supported year
;
; Supported year: 2008-2014 (7 years. Can't do 8 years with bitvectors in EEProm)
;
; OUT:	CC yes, CS no
;		r16 - normalized year (base 2008)
;
;STPossible:
;	ldiw	Z,AVRTimeDateBlk
;	rcall	DSTPossiblefromZ
;	ret

;DSTInit:
;	ldiw	Z,AVRTimeDateBlk
;	rcall	DSTInitfromZ
;	ret

DSTInitfromZ:
	lds		r16,DSTMode		;0..2 are valid values
	cpi		r16,DSTUS
	breq	DSTInitUS		
	cpi		r16,DSTEU
	breq	DSTInitEU		
DSTClrData:
	sts		DSTForwardHr,_0
	sts		DSTForwardDay,_0
	sts		DSTBackDay,_0
	sts		DSTBackMonth,_0
	ret
;
; US:
; DST Forward Month is March, DST Forward Time is 02:00. If match, set time to 3:00
; DST Back Month is November, DST Back Time is 02:00. If match, set to 1:00 (but only once)
; Mark jump back in EEprom. Clear this flag at 02:01
;
; Preserve Z
;
DSTInitUS:
	rcall	DSTPossiblefromZ	;preserves Z
	brcs	DSTClrData
	ldi		r17,DSTUSFORWARDHR
	sts		DSTForwardHr,r17
	ldi		r17,DSTUSBACKMONTH
	sts		DSTBackMonth,r17
	movw	r0,ZL			;preserve
	ldiw	Z,DSTUSData*2	;index in table
	add		ZL,r16			;normalized year returned by DSTPossible
	adc		ZH,_0
	lpm		r16,Z			;get data
	movw	ZL,r0
	mov		r17,r16			;start date
	swap	r17
	andi	r17,0x0f
	sts		DSTForwardDay,r17
	andi	r16,0x0f
	sts		DSTBackDay,r16
	ret
;
; E.U.
; DST Forward Month is March, DST Forward Time is 01:00. If match, set time to 2:00
; DST Back Month is November, DST Back Time is 02:00. If match, set to 1:00 (but only once)
; Mark jump back in EEprom. Clear this flag at 02:01
;
; Preserve Z
;
DSTInitEU:
	rcall	DSTPossiblefromZ	;preserves Z
	brcs	DSTClrData
	sts		DSTForwardHr,_1	;DSTEUFORWARDHR
	ldi		r17,DSTEUBACKMONTH
	sts		DSTBackMonth,r17
	movw	r0,ZL			;preserve
	ldiw	Z,DSTEUData		;index in table
	add		ZL,r16			;normalized year returned by DSTPossible
	adc		ZH,_0
	lpm		r16,Z			;get data
	movw	ZL,r0
	mov		r17,r16			;start date
	swap	r17
	andi	r17,0x0f
	subi	r17,-25
	sts		DSTForwardDay,r17
	andi	r16,0x0f
	subi	r16,-25
	sts		DSTBackDay,r16
	ret
;
; Was Daylight Savings Time Forward applied to the current year?
;
; OUT: r16 flag value
; USED: r17
;
DSTFWasApplied:
	ldiw	Z,AVRTimeDateBlk
	rjmp	DSTFWasAppliedfromZ
;
; DSTFWasAppliedfromZ
;	Was Daylight Savings Time Forward applied to the year in the time block
;	pointed to by Z 
;
; Preserve Z
;
DSTFWasAppliedfromZ:
	rcall	DSTPossiblefromZ ;returns year in r16 (normalized)
	brcc	DSTFWA100
	clr		r16				;preset
	ret
DSTFWA100:
	mov		r17,r16
	lds		r16,DSTFYears	;flags bitvector	
	rjmp	DSTWA200
;
; Was Daylight Savings Time Backward applied to the current year?
;
; OUT: r16 flag value
; USED: r17
;
; Preserve Z
;
DSTBWasApplied:
	ldiw	Z,AVRTimeDateBlk
	rjmp	DSTBWasAppliedfromZ	;Preserves Z

;
; DSTBWasAppliedfromZ
;	Was Daylight Savings Time backward applied to the year in the time block
;	pointed to by Z 
;
; Preserve Z
;
DSTBWasAppliedfromZ:
	rcall	DSTPossiblefromZ ;returns year in r16 (normalized)
	brcc	DSTBWA100
	clr		r16				;preset
	ret
DSTBWA100:
	mov		r17,r16
	lds		r16,DSTBYears	;flags bitvector	
;
; common tail
;
DSTWA200:
	tst		r17				;done?
	breq	DSTWA300		;yes, exit
	dec		r17				;no, continue
	lsr		r16				;start rotate relevant bit to bit 0
	rjmp	DSTWA200
DSTWA300:
	andi	r16,0b00000001	;clear all other years
	ret
;
; Mark Daylight Savings Time Forward applied to the current year
;
; USED: r16, r17
;
DSTFMarkApplied:
	ldiw	Z,AVRTimeDateBlk
	rjmp	DSTFMarkAppliedfromZ
;
; DSTFMarkAppliedfromZ:
; Mark Daylight Savings Time Forward applied to the year in the time block
;	pointed to by Z 
;
; USED: r16, r17
;
; Z not necessarily preserved
;
DSTFMarkAppliedfromZ:
	rcall	DSTPossiblefromZ ;returns year in r16 (normalized)
	brcc	DSTFMA100
	clr		r16				;preset
	ret
DSTFMA100:
	mov		r17,r16
	rcall	DSTMakeMask
	lds		r17,DSTFYears	;current Forward flags bitvector	
	or		r17,r16			;set new flag
	sts		DSTFYears,r17
	mov		r16,r17
	CALL	UpdateDSTFVec_EEprom
#if SERIALIO
	CALL	DSTReportAction	;not completely correct since we may set the Mark
							;without applying the fix (User Time Change)
#endif
	ret
;
; DSTBMarkApplied --
;	Mark Daylight Savings Time Backward applied to the current year
;
; USED: r16, r17
;
DSTBMarkApplied:
	ldiw	Z,AVRTimeDateBlk
	rjmp	DSTBMarkAppliedfromZ
;
; DSTBMarkAppliedfromZ:
; Mark Daylight Savings Time Backward applied to the year in the time block
;	pointed to by Z 
;
; USED: r16, r17
;
; Z not necessarily preserved
;
DSTBMarkAppliedfromZ:
	rcall	DSTPossiblefromZ ;returns year in r16 (normalized)
	brcc	DSTBMA100
	clr		r16				;preset
	ret
DSTBMA100:
	mov		r17,r16
	rcall	DSTMakeMask
	lds		r17,DSTBYears	;current Forward flags bitvector	
	or		r17,r16			;set new flag
	sts		DSTBYears,r17
	mov		r16,r17
	CALL	UpdateDSTBVec_EEprom
#if SERIALIO
	CALL	DSTReportAction	;not completely correct since we may set the Mark
							;without applying the fix (User Time Change)
#endif
	ret
;
; DSTMakeMask --
;	Build the bit mask corresponsing to the normalized year
;
; IN:	r17 normalized year
; OUT:	r16 bitvector mask
;
DSTMakeMask:
	mov		r16,_1			;start with year 0 (2008)
DSTMM100:
	tst		r17				;done?
	breq	DSTMM200		;yes
	dec		r17				;no continue
	lsl		r16				;start rotate to relevant bit
	rjmp	DSTMM100
DSTMM200:
	ret
;
; DSTGPSInWindow --
;	Check if the UTC Year, Month and Day are in the DST window
;
; Need to re-init to make sure proper data are used. Since
; the UTC date will become the AVR date (if different) this is ok.
;
; Also, return FALSE if we are in the last hour of the window since
; we do NOT need a correction in that case
;
; OUT: CS out of range	CC in range
;
DSTGPSInWindow:
DSTGPSInit:
	ldiw	Z,UTCTimeDateBlk
	rcall	DSTInitfromZ		;preserves Z
	rcall	DSTInWindowfromZ	;preserves Z
	brcs	DSTDW100			;outside of window. return FALSE
;
; inside the window. Check for last hr.
;
	lds		r0,GPSLastHrFlag	;were we in the last hour?
	tst		r0
	breq	DSTGPSIL100			;no, return TRUE
;
; We're going to return FALSE (not in window) even though we
; actually are in the last hour of the window. This is fine
; for GPS purposes (we don't want to advance the time outside of
; the DST window) but as a result, the Backward bitvector won't be
; set. Time will be corrected at 2:00AM (set to 1:00) and the
; backward bitvector will be set then, but time will be off until corrected
; by GPS (or the user)
; To prevent this we need to set the Backward bitvector now (just once
; of course), under the assumption that the new GPS time will
; be propagated to the AVR.
; Note that this is only useful if we turn on GPS (or the clock) while
; in the last hour of the window. If we transition into it, the clock
; will fix at 2:00 AM back to 1:00AM, equal to GPS time
;
; This code also assumes that the DST sub-system was initialized
; with UTC data, which we did above.
;
	rcall	DSTBWasAppliedfromZ	;was B flag applied? Preserves Z
	tst		r16	
	brne	DSTDW100			;yes, return FALSE
	ldiw	Z,UTCTimeDateBlk
	rcall	DSTBMarkAppliedfromZ ;set B flag
	rjmp	DSTDW100			;return false
DSTGPSIL100:
	clc
	ret
;
; DSTInWindowfromZ --
;	Check if the Year, Month and Day in the time block
;	pointed to by Z are in the DST window
;
; OUT: CS out of range	CC in range
;
; Preserves Z
;
DSTInWindowfromZ:
	rcall	DSTPossiblefromZ	;preserves Z
	brcc	DSTDW200
DSTDW100:						;return FALSE (out of range)
	sec
	ret
DSTDW200:
	ldd		r16,pMonth
	cpi		r16,DSTFORWARDMONTH
	brlo	DSTDW100		;out of range
	breq	DSTDW300		;first month in window
	lds		r17,DSTBackMonth
	cp		r16,r17
	brlo	DSTDW800		;in range, return true (no carry)
	brne	DSTDW100		;not last month, then out of range
;
; Last month of window. Check Day
;
	ldd		r16,pDay
	lds		r17,DSTBackDay
	cp		r16,r17
	brlo	DSTDW800		;in range, return true (no carry)
	brne	DSTDW100		;not last day, then out of range
;
; last day of window. Are we BEFORE the transition hr
;
	ldd		r16,pHrs
	cpi		r16,DSTBACKHOUR
	brlo	DSTDW400		;yes, return true (no carry)
;
; exit FALSE
;
	sec
	ret						;no but last day so we're done
;
; last day of window, before transition hr
;
; Are we in the LAST HOUR before the transition hr.
; This is relevant for the GPS parsing/DST combination
; since in that last hour, we should NOT advance time 1 hr
;
DSTDW400:
	sts		GPSLastHrFlag,_0	;clear special flag
	cpi		r16,DSTBACKHOUR-1	;are we inside the last hour window?
	brlo	DSTDW800			;no, just return TRUE
;
; Inside the LAST HOUR of the window.
;
	sts		GPSLastHrFlag,_1	;set special flag
	rjmp	DSTDW800			;return TRUE
;
; First Month in window. Check Day
;
DSTDW300:
	ldd		r16,pDay
	lds		r17,DSTForwardDay
	cp		r16,r17
	brlo	DSTDW100		;out of range
	brne	DSTDW800		;not first Day
;
; First day of window. Are we after the transistion hr
;
	ldd		r16,pHrs
	lds		r17,DSTForwardHr
	cp		r16,r17
	brsh	DSTDW800		;yes, return true (no carry)
	sec
	ret						;no but first day so we're done
;
; Month and Day in window
;
DSTDW800:
	clc
	ret
;
; DSTPostWindowfromZ --
;	Check if the Year, Month and Day in the time block
;	pointed to by Z are post the DST windows
;
; OUT: CS out of range	CC in range
;
DSTPostWindowfromZ:
	rcall	DSTPossiblefromZ	;preserves Z
	brcc	DSTPW200
DSTPW100:
	sec
	ret
DSTPW200:
	ldd		r16,pMonth
	lds		r17,DSTBackMonth
	cp		r17,r16			;reverse compare
	brlo	DSTPW800		;post range, return true (no carry)
	brne	DSTPW100		;not last month, then out of range
;
; Last month of window. Check Day
;
	ldd		r16,pDay
	lds		r17,DSTBackDay
	cp		r17,r16			;reverse compare
	brlo	DSTPW800		;post range, return true (no carry)
	brne	DSTPW100		;not last day, then out of range
;
; last day of window. Are we after the transition hr
;
	ldd		r16,pHrs
	cpi		r16,DSTBACKHOUR
	brsh	DSTPW800		;yes, return true (no carry)
	sec
	ret						;no but last day so we're done
;
; Month and Day in window
;
DSTPW800:
	clc
	ret
